# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: relationalai/lqp/v1/logic.proto
# Protobuf Python Version: 4.25.6
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1frelationalai/lqp/v1/logic.proto\x12\x13relationalai.lqp.v1\"\x80\x01\n\x0b\x44\x65\x63laration\x12,\n\x03\x64\x65\x66\x18\x01 \x01(\x0b\x32\x18.relationalai.lqp.v1.DefH\x00R\x03\x64\x65\x66\x12/\n\x04loop\x18\x02 \x01(\x0b\x32\x19.relationalai.lqp.v1.LoopH\x00R\x04loopB\x12\n\x10\x64\x65\x63laration_type\"\xa6\x01\n\x03\x44\x65\x66\x12\x33\n\x04name\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\x04name\x12\x34\n\x04\x62ody\x18\x02 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x62ody\x12\x34\n\x05\x61ttrs\x18\x03 \x03(\x0b\x32\x1e.relationalai.lqp.v1.AttributeR\x05\x61ttrs\"\xaf\x01\n\x04Loop\x12\x41\n\x0ctemporal_var\x18\x01 \x01(\x0b\x32\x1e.relationalai.lqp.v1.LoopIndexR\x0btemporalVar\x12.\n\x05inits\x18\x02 \x03(\x0b\x32\x18.relationalai.lqp.v1.DefR\x05inits\x12\x34\n\x04\x62ody\x18\x03 \x03(\x0b\x32 .relationalai.lqp.v1.DeclarationR\x04\x62ody\"o\n\x0b\x41\x62straction\x12,\n\x04vars\x18\x01 \x03(\x0b\x32\x18.relationalai.lqp.v1.VarR\x04vars\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x05value\"\x8a\x05\n\x07\x46ormula\x12\x35\n\x06\x65xists\x18\x01 \x01(\x0b\x32\x1b.relationalai.lqp.v1.ExistsH\x00R\x06\x65xists\x12\x35\n\x06reduce\x18\x02 \x01(\x0b\x32\x1b.relationalai.lqp.v1.ReduceH\x00R\x06reduce\x12\x44\n\x0b\x63onjunction\x18\x03 \x01(\x0b\x32 .relationalai.lqp.v1.ConjunctionH\x00R\x0b\x63onjunction\x12\x44\n\x0b\x64isjunction\x18\x04 \x01(\x0b\x32 .relationalai.lqp.v1.DisjunctionH\x00R\x0b\x64isjunction\x12,\n\x03not\x18\x05 \x01(\x0b\x32\x18.relationalai.lqp.v1.NotH\x00R\x03not\x12,\n\x03\x66\x66i\x18\x06 \x01(\x0b\x32\x18.relationalai.lqp.v1.FFIH\x00R\x03\x66\x66i\x12/\n\x04\x61tom\x18\x07 \x01(\x0b\x32\x19.relationalai.lqp.v1.AtomH\x00R\x04\x61tom\x12\x35\n\x06pragma\x18\x08 \x01(\x0b\x32\x1b.relationalai.lqp.v1.PragmaH\x00R\x06pragma\x12>\n\tprimitive\x18\t \x01(\x0b\x32\x1e.relationalai.lqp.v1.PrimitiveH\x00R\tprimitive\x12\x36\n\x08true_val\x18\n \x01(\x0b\x32\x19.relationalai.lqp.v1.TrueH\x00R\x07trueVal\x12\x39\n\tfalse_val\x18\x0b \x01(\x0b\x32\x1a.relationalai.lqp.v1.FalseH\x00R\x08\x66\x61lseValB\x0e\n\x0c\x66ormula_type\"j\n\x06\x45xists\x12,\n\x04vars\x18\x01 \x03(\x0b\x32\x18.relationalai.lqp.v1.VarR\x04vars\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x05value\"\xa1\x01\n\x06Reduce\x12\x30\n\x02op\x18\x01 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x02op\x12\x34\n\x04\x62ody\x18\x02 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x62ody\x12/\n\x05terms\x18\x03 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"?\n\x0b\x43onjunction\x12\x30\n\x04\x61rgs\x18\x01 \x03(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x04\x61rgs\"?\n\x0b\x44isjunction\x12\x30\n\x04\x61rgs\x18\x01 \x03(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x04\x61rgs\"5\n\x03Not\x12.\n\x03\x61rg\x18\x01 \x01(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x03\x61rg\"\x80\x01\n\x03\x46\x46I\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x34\n\x04\x61rgs\x18\x02 \x03(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x61rgs\x12/\n\x05terms\x18\x03 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"l\n\x04\x41tom\x12\x33\n\x04name\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\x04name\x12/\n\x05terms\x18\x02 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"M\n\x06Pragma\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12/\n\x05terms\x18\x02 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"P\n\tPrimitive\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12/\n\x05terms\x18\x02 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"\x06\n\x04True\"\x07\n\x05\x46\x61lse\"~\n\x04Term\x12,\n\x03var\x18\x01 \x01(\x0b\x32\x18.relationalai.lqp.v1.VarH\x00R\x03var\x12;\n\x08\x63onstant\x18\x02 \x01(\x0b\x32\x1d.relationalai.lqp.v1.ConstantH\x00R\x08\x63onstantB\x0b\n\tterm_type\"Q\n\x03Var\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x36\n\x04type\x18\x02 \x01(\x0e\x32\".relationalai.lqp.v1.PrimitiveTypeR\x04type\"E\n\x08\x43onstant\x12\x39\n\x05value\x18\x01 \x01(\x0b\x32#.relationalai.lqp.v1.PrimitiveValueR\x05value\"R\n\tAttribute\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x31\n\x04\x61rgs\x18\x02 \x03(\x0b\x32\x1d.relationalai.lqp.v1.ConstantR\x04\x61rgs\"<\n\nRelationId\x12\x15\n\x06id_low\x18\x01 \x01(\x06R\x05idLow\x12\x17\n\x07id_high\x18\x02 \x01(\x06R\x06idHigh\"\x1f\n\tLoopIndex\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\"\x80\x01\n\x0ePrimitiveValue\x12#\n\x0cstring_value\x18\x01 \x01(\tH\x00R\x0bstringValue\x12\x1d\n\tint_value\x18\x02 \x01(\x03H\x00R\x08intValue\x12!\n\x0b\x66loat_value\x18\x03 \x01(\x01H\x00R\nfloatValueB\x07\n\x05value*|\n\rPrimitiveType\x12\x1e\n\x1aPRIMITIVE_TYPE_UNSPECIFIED\x10\x00\x12\x19\n\x15PRIMITIVE_TYPE_STRING\x10\x01\x12\x16\n\x12PRIMITIVE_TYPE_INT\x10\x02\x12\x18\n\x14PRIMITIVE_TYPE_FLOAT\x10\x03\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'relationalai.lqp.v1.logic_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_PRIMITIVETYPE']._serialized_start=2768
  _globals['_PRIMITIVETYPE']._serialized_end=2892
  _globals['_DECLARATION']._serialized_start=57
  _globals['_DECLARATION']._serialized_end=185
  _globals['_DEF']._serialized_start=188
  _globals['_DEF']._serialized_end=354
  _globals['_LOOP']._serialized_start=357
  _globals['_LOOP']._serialized_end=532
  _globals['_ABSTRACTION']._serialized_start=534
  _globals['_ABSTRACTION']._serialized_end=645
  _globals['_FORMULA']._serialized_start=648
  _globals['_FORMULA']._serialized_end=1298
  _globals['_EXISTS']._serialized_start=1300
  _globals['_EXISTS']._serialized_end=1406
  _globals['_REDUCE']._serialized_start=1409
  _globals['_REDUCE']._serialized_end=1570
  _globals['_CONJUNCTION']._serialized_start=1572
  _globals['_CONJUNCTION']._serialized_end=1635
  _globals['_DISJUNCTION']._serialized_start=1637
  _globals['_DISJUNCTION']._serialized_end=1700
  _globals['_NOT']._serialized_start=1702
  _globals['_NOT']._serialized_end=1755
  _globals['_FFI']._serialized_start=1758
  _globals['_FFI']._serialized_end=1886
  _globals['_ATOM']._serialized_start=1888
  _globals['_ATOM']._serialized_end=1996
  _globals['_PRAGMA']._serialized_start=1998
  _globals['_PRAGMA']._serialized_end=2075
  _globals['_PRIMITIVE']._serialized_start=2077
  _globals['_PRIMITIVE']._serialized_end=2157
  _globals['_TRUE']._serialized_start=2159
  _globals['_TRUE']._serialized_end=2165
  _globals['_FALSE']._serialized_start=2167
  _globals['_FALSE']._serialized_end=2174
  _globals['_TERM']._serialized_start=2176
  _globals['_TERM']._serialized_end=2302
  _globals['_VAR']._serialized_start=2304
  _globals['_VAR']._serialized_end=2385
  _globals['_CONSTANT']._serialized_start=2387
  _globals['_CONSTANT']._serialized_end=2456
  _globals['_ATTRIBUTE']._serialized_start=2458
  _globals['_ATTRIBUTE']._serialized_end=2540
  _globals['_RELATIONID']._serialized_start=2542
  _globals['_RELATIONID']._serialized_end=2602
  _globals['_LOOPINDEX']._serialized_start=2604
  _globals['_LOOPINDEX']._serialized_end=2635
  _globals['_PRIMITIVEVALUE']._serialized_start=2638
  _globals['_PRIMITIVEVALUE']._serialized_end=2766
# @@protoc_insertion_point(module_scope)
