# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: relationalai/lqp/v1/fragments.proto
# Protobuf Python Version: 4.25.6
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from relationalai.lqp.v1 import logic_pb2 as relationalai_dot_lqp_dot_v1_dot_logic__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#relationalai/lqp/v1/fragments.proto\x12\x13relationalai.lqp.v1\x1a\x1frelationalai/lqp/v1/logic.proto\"\x81\x01\n\x08\x46ragment\x12/\n\x02id\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.FragmentIdR\x02id\x12\x44\n\x0c\x64\x65\x63larations\x18\x02 \x03(\x0b\x32 .relationalai.lqp.v1.DeclarationR\x0c\x64\x65\x63larations\"\x1c\n\nFragmentId\x12\x0e\n\x02id\x18\x01 \x01(\x0cR\x02idb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'relationalai.lqp.v1.fragments_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_FRAGMENT']._serialized_start=94
  _globals['_FRAGMENT']._serialized_end=223
  _globals['_FRAGMENTID']._serialized_start=225
  _globals['_FRAGMENTID']._serialized_end=253
# @@protoc_insertion_point(module_scope)
